import qualified Control.Exception as Exception
import Control.Monad

import LazyContract

elim1 :: Either a b -> a
elim1 (Left x) = x

elim2 :: Either a b -> b
elim2 (Right x) = x

eqInt :: Int -> Int -> Bool
eqInt = (==)

ex1 abcd = CasePair abcd 
           (\ (ab,cd) -> CasePair ab 
                         (\ (a,b) -> CasePair cd 
                                     (\ (c,d) -> Pred "a == c"
                                                   (do x <- demand a 
                                                       y <- demand c
                                                       return $ x `eqInt` y))))

test1 = let p = assert (Prop ex1) (unitLocs "pair p in test1") ((1,2),(3,4))
        in map show [snd (snd p), fst (snd p), snd (fst p), fst (fst p)]

test2 = let p = assert (Prop ex1) (unitLocs "pair p in test2") ((1,2),(1,4))
        in map show [snd (snd p), fst (snd p), snd (fst p), fst (fst p)]

ex2 abcd = CasePair abcd
             (\ (ab,cd) -> CasePair ab
                             (\ (a,b) -> CaseSum cd
                                           (\c -> Pred "a == c"
                                                    (do x <- demand a
                                                        y <- demand c
                                                        return $ x `eqInt` y))
                                           (\d -> Pred "True" (return True))))

test3 = let p = assert (Prop ex2) (unitLocs "pair p in test3") ((1,2), Left 3)
        in map show [snd (fst p), fst (fst p), elim1 (snd p)]

test4 = let p = assert (Prop ex2) (unitLocs "pair p in test4") ((1,2), Left 1)
            p' = assert (Prop ex2) (unitLocs "pair p' in test4") ((1,2), Right 3)
        in map show [snd (fst p), fst (fst p), elim1 (snd p),
                     snd (fst p), fst (fst p), elim2 (snd p')]

ex3 abcd = CasePair abcd
             (\ (ab,cd) -> CaseSum ab
                             (\a -> CaseSum cd
                                      (\c -> Pred "a == c"
                                               (do x <- demand a
                                                   y <- demand c
                                                   return $ x `eqInt` y))
                                      (\d -> Pred "True" $ return True))
                             (\b -> Pred "True" $ return True))

test5 = let p = assert (Prop ex3) (unitLocs "pair p in test5") (Left 1, Left 2)
        in map show [elim1 (fst p), elim1 (snd p)]

test6 = let p1 = assert (Prop ex3) (unitLocs "pair p1 in test6") (Left 1, Right 2)
            p2 = assert (Prop ex3) (unitLocs "pair p2 in test6") (Right 1, Left 2)
        in map show [elim1 (fst p1), elim2 (snd p1), elim2 (fst p2), elim1 (snd p2)]

ex4 =
  let post b b' = CasePair b' (\ (x,y) -> Pred "POST: x == function-argument"
                                            (do x' <- demand x
                                                arg <- demand b
                                                return $ x' `eqInt` arg))
      pre b = Pred "PRE: function-argument >= 0"
                (do x <- demand b
                    return (x >= 0))
  in Function (Prop pre) (\b -> Prop (post b))

test7 = let f = assert ex4 (unitLocs "<0>") (Fun (\lx x -> (x,0)))
        in map show [fst (app f (unitLocs "<1>") 0), 
                     fst (app f (unitLocs "<2>") 1), 
                     snd (app f (unitLocs "<3>") (-1)), 
                     fst (app f (unitLocs "<4>") (-1))]

test8 = let f = assert ex4 (unitLocs "<5>") (Fun (\lx x -> (1,0)))
        in map show [app f (unitLocs "<6>") 0]

test9 = let f = assert ex4 (unitLocs "<7>") (Fun (\lx x -> (x-1,0)))
        in map show [snd (app f (unitLocs "<8>") 3), 
                     fst (app f (unitLocs "<9>") 3)]

ex5 = let post b = Pred "POST" (do x <- demand b
                                   return $ x > 0)
          pre _ = Pred "PRE" (return True)
      in Function (Prop pre) (\_ -> Prop post)

test10 = let f = assert ex5 (unitLocs "<10>") (Fun (\lx x -> x))
         in map show [app f (unitLocs "<11>") 1,
                      app f (unitLocs "<12>") 0]

ex6 = let post b b' = Pred "POST" (do x <- demand b
                                      x' <- demand b'
                                      return $ x `eqInt` x')
          pre _ = Pred "PRE" (return True)
      in Function (Prop pre) (\b -> Prop $ post b)

test11 = let f = assert ex6 (unitLocs "<13>") (Fun (\lx x -> x + 1))
         in map show [app f (unitLocs "<14>") 1]

ex7 =
  let post b = CasePair b (\ (x,y) -> Pred "POST"
                                            (do x' <- demand x
                                                return (x' `eqInt` 0)))
      pre _ = Pred "PRE" (return True)
  in Function  (Prop pre) (\_ -> Prop post)

test12 = let f = assert ex7 (unitLocs "<15>") (Fun (\lx x -> (x,1)))
         in map show [fst (app f (unitLocs "<16>") 1)]

ex8 = let p b = CasePair b (\ (x,y) -> Pred "first component == 0"
                                            (do x' <- demand x
                                                return (x' `eqInt` 0)))
      in Prop p

test13 = let p = assert ex8 (unitLocs "<17>") (1,2)
         in map show [fst p]

test14 =
  let contract = Pair (Prop $ \b -> Pred "x == 0" 
                                      $ demand b >>= \x -> return $ x `eqInt` 0)
                      (Prop $ \_ -> Pred "True" $ return True)
      p = assert contract (unitLocs "<18>")
  in map show [fst (p (0,3)), snd (p (1,2)), fst (p (1,2))]

test15 =
  let contract = Sum (Prop $ \b -> Pred "x == 0"
                                     $ demand b >>= \x -> return $ x `eqInt` 0)
                     (Prop $ \_ -> Pred "True" $ return True)
      p = assert contract (unitLocs "<19>")
  in map show [p (Left 0), p (Right 3), p (Left 1)]

ex9 =
  let c1 b = CasePair b (\ (x,y) -> Pred "x == 0"
                                            (do x' <- demand x
                                                return (x' `eqInt` 0)))
      c2 b = CasePair b (\ (x,y) -> Pred "y == 1"
                                            (do y' <- demand y
                                                return (y' `eqInt` 1)))
   in And (Prop c1) (Prop c2)
      
test16 =
  let p = assert ex9 (unitLocs "<20>") (0,1)
  in map show [p]

test17 =
  let p = assert ex9 (unitLocs "<21>") (1,1)
  in map show [snd p, fst p]

test18 =
  let p = assert ex9 (unitLocs "<22>") (0,0)
  in map show [fst p, snd p]

type Test = (String, [String], Maybe String)

allTests :: [Test]
allTests = [("test1", test1, Just ("predicate a == c failed, the expression " ++
                                   "\"pair p in test1\" is to blame")),
            ("test2", test2, Nothing),
            ("test3", test3, Just ("predicate a == c failed, the expression " ++
                                   "\"pair p in test3\" is to blame")),
            ("test4", test4, Nothing),
            ("test5", test5, Just ("predicate a == c failed, the expression " ++
                                   "\"pair p in test5\" is to blame")),
            ("test6", test6, Nothing),
            ("test7", test7, Just ("predicate PRE: function-argument >= 0 failed, " ++
                                   "the expression \"<4>\" is to blame")),
            ("test8", test8, Nothing),
            ("test9", test9, Just ("predicate POST: x == function-argument " ++
                                   "failed, the expression \"<7>\" is to blame")),
            ("test10", test10, Just ("predicate POST failed, the expression " ++
                                     "\"<10>\" is to blame")),
            ("test11", test11, Just ("predicate POST failed, the expression " ++
                                     "\"<13>\" is to blame")),
            ("test12", test12, Just ("predicate POST failed, the expression " ++
                                     "\"<15>\" is to blame")),
            ("test13", test13, Just ("predicate first component == 0 failed, " ++
                                     "the expression \"<17>\" is to blame")),
            ("test14", test14, Just ("predicate x == 0 failed, the expression " ++
                                     "\"<18>\" is to blame")),
            ("test15", test15, Just ("predicate x == 0 failed, the expression " ++
                                     "\"<19>\" is to blame")),
            ("test16", test16, Nothing),
            ("test17", test17, Just ("predicate x == 0 failed, the expression " ++
                                     "\"<21>\" is to blame")),
            ("test18", test18, Just ("predicate y == 1 failed, the expression " ++
                                     "\"<22>\" is to blame"))
           ]
           
data TestResult = TestOK String | TestFail String String

isTestFail (TestOK _) = False
isTestFail (TestFail _ _) = True

writeDevNull = writeFile "/dev/null"

runTest :: Test -> IO TestResult
runTest (name, [], _) = return $ TestFail name "test returned the empty list"
runTest (name, l, exp) =
  let (oks, last) = splitAt (length l - 1) l
  in do res1 <- runWithoutContractViolation oks
        case res1 of
          TestOK _ -> 
            case exp of
              Nothing -> runWithoutContractViolation last
              Just s -> runWithContractViolation s last
          TestFail _ _ -> return res1
  where 
  runWithoutContractViolation l =
    Exception.catchJust isContractViolation 
      (do mapM_ writeDevNull l
          return $ TestOK name)
      (\s -> return $ TestFail name ("unexpected contract violation: " 
                                     ++ show s))
  runWithContractViolation s l =
    Exception.catchJust isContractViolation 
      (do mapM_ writeDevNull l
          return $ TestFail name ("unexpected pass"))
      (\s' -> return $ if s == s' 
                          then TestOK name
                          else TestFail name
                                 ("wrong contract violation, expected " 
                                  ++ show s ++ ", occurred " ++ show s'))

runTests :: [Test] -> IO ()
runTests l =
  do results <- mapM runTest l
     let failures = filter isTestFail results
         nfailures = length failures
         nok = ntests - nfailures
         ntests = length results
     putStrLn (show ntests ++ " tests: " ++ show nfailures ++ " failures, " ++
               show nok ++ " ok")
     when (nfailures > 0) $
       do putStrLn "Failures:"
          mapM_ (\ (TestFail name s) -> putStrLn (" * Test " ++ name ++ ": " 
                                                  ++ s)) failures
   
runAllTests = runTests allTests

main = runAllTests
