{-#  OPTIONS -fglasgow-exts  #-}
module TypeMap where

data Addr :: (* -> *) -> * -> * -> * where
  AtTop :: f a -> Addr f a (f a)
  AtLft :: Addr f a c -> Addr f (Either a b) c
  AtRgt :: Addr f b c -> Addr f (Either a b) c
  AtFst :: Addr f a c -> Addr f (a, b) c
  AtSnd :: Addr f b c -> Addr f (a, b) c

data Map :: (* -> *) -> * -> * where
  ValNode  :: f a -> Map f a -> Map f a	       -- INV: no two successive ValNodes
  PairNode :: Map f a -> Map f b -> Map f (a,b)
  SumNode  :: Map f a -> Map f b -> Map f (Either a b)
  PlugNode :: Map f a

emptyMap = PlugNode

-- lookup an address in the map
lkup :: Addr f t (f a) -> Map f t -> Maybe (f a)
lkup (AtTop x) (ValNode p m) = Just p
lkup (AtLft x) (ValNode p m)  = lkup (AtLft x) m
lkup (AtLft x) (SumNode ma mb)  = lkup x ma
lkup (AtRgt x) (ValNode p m)  = lkup (AtRgt x) m
lkup (AtRgt x) (SumNode ma mb)  = lkup x mb
lkup (AtFst x) (ValNode p m) = lkup (AtFst x) m
lkup (AtFst x) (PairNode ma mb) = lkup x ma
lkup (AtSnd x) (ValNode p m) = lkup (AtSnd x) m
lkup (AtSnd x) (PairNode ma mb) = lkup x mb
lkup _ _ = Nothing

-- insert an address string into the map
insert :: Addr f t (f a) -> Map f t -> Map f t
insert (AtTop x) (PlugNode) = (ValNode x PlugNode)
-- 
insert (AtLft x) (ValNode v m) = ValNode v (insert (AtLft x) m)
insert (AtLft x) (PlugNode) = (SumNode (insert x PlugNode) PlugNode)
insert (AtLft x) (SumNode ma mb) = SumNode (insert x ma) mb
-- 
insert (AtRgt x) (ValNode v m) = ValNode v (insert (AtRgt x) m)
insert (AtRgt x) (PlugNode) = (SumNode PlugNode (insert x PlugNode))
insert (AtRgt x) (SumNode ma mb) = SumNode ma (insert x mb)
-- 
insert (AtFst x) (ValNode v m) = ValNode v (insert (AtFst x) m)
insert (AtFst x) (PlugNode) = (PairNode (insert x PlugNode) PlugNode)
insert (AtFst x) (PairNode ma mb) = PairNode (insert x ma) mb
-- 
insert (AtSnd x) (ValNode v m) = ValNode v (insert (AtSnd x) m)
insert (AtSnd x) (PlugNode) = (PairNode PlugNode (insert x PlugNode))
insert (AtSnd x) (PairNode ma mb) = PairNode ma (insert x mb)

-- create a new address
addLft :: Addr f t (f (Either a b)) -> f a -> Addr f t (f a)
addLft (AtTop x) v = (AtLft (AtTop v))
addLft (AtLft a) v = AtLft (addLft a v)
addLft (AtRgt a) v = AtRgt (addLft a v)
addLft (AtFst a) v = AtFst (addLft a v)
addLft (AtSnd a) v = AtSnd (addLft a v)

addRgt :: Addr f t (f (Either a b)) -> f b -> Addr f t (f b)
addRgt (AtTop x) v = (AtRgt (AtTop v))
addRgt (AtLft a) v = AtLft (addRgt a v)
addRgt (AtRgt a) v = AtRgt (addRgt a v)
addRgt (AtFst a) v = AtFst (addRgt a v)
addRgt (AtSnd a) v = AtSnd (addRgt a v)

addFst :: Addr f t (f (a, b)) -> f a -> Addr f t (f a)
addFst (AtTop x) v = (AtFst (AtTop v))
addFst (AtLft a) v = AtLft (addFst a v)
addFst (AtRgt a) v = AtRgt (addFst a v)
addFst (AtFst a) v = AtFst (addFst a v)
addFst (AtSnd a) v = AtSnd (addFst a v)

addSnd :: Addr f t (f (a, b)) -> f b -> Addr f t (f b)
addSnd (AtTop x) v = (AtSnd (AtTop v))
addSnd (AtLft a) v = AtLft (addSnd a v)
addSnd (AtRgt a) v = AtRgt (addSnd a v)
addSnd (AtFst a) v = AtFst (addSnd a v)
addSnd (AtSnd a) v = AtSnd (addSnd a v)

